# -*- coding: utf-8 -*-
from __future__ import (unicode_literals, print_function, division,
                        absolute_import)

import os
import asyncore
from smtpd import SMTPServer
import tempfile

import click

from .store import Store
from .pop3 import Server as Pop3Server


def parse_address(address):
    parts = address.split(':')
    return ':'.join(parts[:-1]), int(parts[-1])


def validate_address(ctx, param, value):
    try:
        return parse_address(value)
    except ValueError:
        raise click.BadParameter('address needs to be in format host:port')


def print_version(ctx, param, value):
    from . import __version__
    if not value or ctx.resilient_parsing:
        return
    click.echo('Version ' + __version__)
    ctx.exit()


@click.command()
@click.option('--version', is_flag=True, callback=print_version, expose_value=False, is_eager=True, help='Print version and exit')
@click.option('-m', '--mode', type=click.Choice(['smtp', 'pop3', 'both']), default='both', help='Run smtp, pop3 or both (default)')
@click.option('-sa', '--smtp-address', default='localhost:5050', help='Address to run the SMTP server on. Defaults to localhost:5050', callback=validate_address)
@click.option('-pa', '--pop3-address', default='localhost:5051', help='Address to run the POP3 server on. Defaults to localhost:5051', callback=validate_address)
@click.option('-pu', '--pop3-user', default=None, help='Username for the POP3 server (default: <any>)')
@click.option('-pp', '--pop3-password', default=None, help='Password for the POP3 server (default: <any>)')
@click.argument('directory', type=click.Path(resolve_path=True), required=False)
def cli(mode, smtp_address, pop3_address, pop3_user, pop3_password, directory):
    """
    Runs an SMTP server, POP3 server or both based on a directory.

    When no directory is supplied, a temporary directory will be created and
    used. The POP3 server accepts any username and password combination by
    default.

    """
    if directory is None:
        directory = tempfile.mkdtemp()
        click.echo('Running from directory %s' % directory)
    else:
        if not os.path.exists(directory):
            os.makedirs(directory)
    store = Store(directory)
    store.load()
    if mode in ('smtp', 'both'):
        Smtp(smtp_address, None, store=store)
    if mode in ('pop3', 'both'):
        Pop3Server(pop3_address, store, pop3_user, pop3_password)
    try:
        asyncore.loop()
    except KeyboardInterrupt:
        pass


class Smtp(SMTPServer):

    def __init__(self, *args, **kwargs):
        self.store = kwargs.pop('store')
        SMTPServer.__init__(self, *args, **kwargs)

    def process_message(self, peer, mailfrom, rcpttos, data):
        self.store.save(data)



# -*- coding: utf-8 -*-
from __future__ import (unicode_literals, print_function, division,
                        absolute_import)

import asyncore
import asynchat
import socket


class Request(object):

    def __init__(self, command, *args):
        self.command = command
        self.args = args

    @property
    def arg(self):
        return None if not self.args else self.args[0]

    @property
    def message_nr(self):
        arg = self.arg
        return None if arg is None else int(arg)

    @property
    def has_args(self):
        return self.args is not None and len(self.args) > 0

    @property
    def merged_arg(self):
        if self.args is None:
            return None
        return ' '.join(self.args)


class Response(object):

    def __init__(self, status, lines=None):
        self.status = status
        self.lines = lines

    @classmethod
    def error(cls, lines=None):
        return cls('-ERR', lines)

    @classmethod
    def ok(cls, lines=None):
        return cls('+OK', lines)

    @classmethod
    def ok_extra(cls, msg, lines=None):
        return cls('+OK %s' % msg, lines)


class Pop3Exception(Exception):
    pass


def file_to_lines(path):
    with open(path, 'rb') as f:
        last = None
        new_line = True
        line = []
        while True:
            current = f.read(1)
            if current == b'':
                # EOF
                yield ''.join(line)
                return
            line.append(current)
            if new_line:
                new_line = False
                if current == b'.':
                    # EOT indicator is .\r\n -- stuff extra dot at new line
                    line.append(b'.')
            if last == b'\r' and current == 'b\n':
                # Done with this line; pass it without \r\n
                yield ''.join(line[:-2])
                line = []
            last = current


class Session(asynchat.async_chat):

    def __init__(self, sock, store, user, password):
        asynchat.async_chat.__init__(self, sock)
        self.store = store
        self.user = user
        self.password = password
        self.buffer = []
        self.set_terminator('\r\n')

    def collect_incoming_data(self, data):
        self.buffer.append(data)

    def clear_buffer(self):
        raw_request = ''.join(self.buffer).strip()
        self.buffer = []
        return raw_request

    def found_terminator(self):
        request_parts = self.clear_buffer().split()
        request = Request(request_parts[0], *request_parts[1:])
        handler = getattr(self, 'do_%s' % request.command.lower(), None)
        response = None
        if handler and callable(handler):
            try:
                response = handler(request)
            except Pop3Exception:
                response = Response.error()
        if not response:
            response = Response.error()
        self.respond(response)

    def do_noop(self, request):
        return Response.ok()

    def do_user(self, request):
        if not self.user:
            # Anything will do when no user was set
            return Response.ok()
        if self.user == request.merged_arg:
            return Response.ok()
        raise Pop3Exception()

    def do_pass(self, request):
        # TODO Begin transaction
        if not self.password:
            # Anything will do when no password was set
            return Response.ok()
        if self.password == request.merged_arg:
            return Response.ok()
        raise Pop3Exception()

    def do_quit(self, request):
        # TODO Commit transaction
        if self.store.delete_marked_messages():
            return Response.ok()
        raise Pop3Exception()

    def do_retr(self, request):
        m = self.store.get(request.message_nr)
        if not m:
            raise Pop3Exception()
        return Response.ok(file_to_lines(m.path))

    def do_capa(self, request):
        capabilities = [n for n in dir(self)
                        if n.startswith('do_') and callable(getattr(self, n))]
        return Response.ok([n[3:].upper() for n in capabilities])

    def do_stat(self, request):
        msg = '%d %d' % (len(self.store), self.store.total_byte_size)
        return Response.ok_extra(msg)

    def do_rset(self, request):
        m = self.store.get(request.message_nr, include_deleted=True)
        if m:
            m.deleted = False
            return Response.ok()
        raise Pop3Exception()

    def do_dele(self, request):
        m = self.store.get(request.message_nr)
        if m:
            m.deleted = True
            return Response.ok()
        raise Pop3Exception()

    def do_list(self, request):
        if not request.has_args:
            return Response.ok(['%d %d' % (m.nr, m.size) for m in self.store])
        m = self.store.get(request.message_nr)
        if m:
            return Response.ok_extra('%d %d' % (m.nr, m.size))
        raise Pop3Exception()

    def do_uidl(self, request):
        if not request.has_args:
            return Response.ok(['%d %s' % (m.nr, m.uid) for m in self.store])
        m = self.store.get(request.message_nr)
        if m:
            return Response.ok_extra('%d %s' % (m.nr, m.uid))
        raise Pop3Exception()

    def respond(self, response):
        self.push((b'%s\r\n' % response.status).encode('ascii'))
        if response.lines:
            for line in response.lines:
                msg = b'%s\r\n' % line
                self.push(msg.encode('ascii'))
            self.push(b'.\r\n'.encode('ascii'))


class Server(asyncore.dispatcher):

    def __init__(self, address, store, user, password):
        asyncore.dispatcher.__init__(self)
        self.store = store
        self.user = user
        self.password = password
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind(address)
        self.listen(5)

    def handle_accept(self):
        pair = self.accept()
        if pair is not None:
            # TODO Handle disconnect
            sock, addr = pair
            sock.send(b'+OK Welcome to Flying Rat\r\n')
            Session(sock, self.store, self.user, self.password)



import os
from setuptools import setup

def read(*paths):
    with open(os.path.join(*paths), 'r') as f:
        return f.read()

setup(
    name='flyingrat',
    version='0.1.0',
    description='Simple mail server for local development',
    long_description=read('README.rst'),
    url='https://github.com/roam/flyingrat/',
    author='Kevin Wetzels',
    author_email='kevin@roam.be',
    license='BSD',
    packages=['flyingrat'],
    include_package_data=True,
    install_requires=[
        'Click'
    ],
    entry_points='''
        [console_scripts]
        flyingrat=flyingrat.cli:cli
    ''',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ],
)



# -*- coding: utf-8 -*-
from __future__ import (unicode_literals, print_function, division,
                        absolute_import)

import os
import errno
import uuid
from datetime import datetime


class Message(object):

    def __init__(self, nr, uid, path, size):
        self.nr = nr
        self.uid = uid
        self.path = path
        self.size = size
        self.deleted = False


class Store(object):

    def __init__(self, directory):
        self.directory = directory
        self.counter = 0
        self.messages = []

    def __len__(self):
        return len(self.non_deleted_messages)

    def __iter__(self):
        for m in self.non_deleted_messages:
            yield m

    @property
    def total_byte_size(self):
        return sum((m.size for m in self.non_deleted_messages))

    @property
    def non_deleted_messages(self):
        return [m for m in self.messages if not m.deleted]

    def load(self):
        self.messages = []
        self.counter = 0
        for filename in os.listdir(self.directory):
            if not filename.endswith('.eml'):
                continue
            uid = self.parse_uid(filename)
            path = os.path.join(self.directory, filename)
            size = os.stat(path).st_size
            self.counter += 1
            self.messages.append(Message(self.counter, uid, path, size))

    def get(self, nr, include_deleted=False):
        messages = self.messages
        if not include_deleted:
            messages = self.non_deleted_messages
        for m in messages:
            if m.nr == nr:
                return m
        return None

    def save(self, data):
        timestamp = datetime.now().strftime('%Y%m%d-%H%M%S')
        uid = uuid.uuid4().hex
        filename = '%s-%s.eml' % (timestamp, uid)
        path = os.path.join(self.directory, filename)
        with open(path, 'wb+') as f:
            f.write(data)
        self.counter += 1
        stat = os.stat(path)
        m = Message(self.counter, uid, path, stat.st_size)
        self.messages.append(m)
        return m

    def parse_uid(self, filename):
        return b'-'.join(filename[0:-len('.eml')].split('-')[1:])

    def delete_marked_messages(self):
        for m in self.messages:
            if m.deleted:
                try:
                    os.unlink(m.path)
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise
                self.messages.remove(m)
        return True

